### Subroutine:    &register()
###
### Function:      Processes data from a basic registration form, including
###                the following fields:
###
###                NAME:      User's full name
###                MAIL:      E-Mail address
###                EMAIL2:    E-Mail address verification (should match EMAIL)
###                USER_ID:   User ID (4-10 printable characters)
###                PASSWORD:  Password (6-10 printable characters)
###                PASSWORD2: Password verification (should match PASSWORD)
###                VERIFY:    Verification field (such as Mother's maiden name)
###                    
###
### Usage:         &register;
###
### Variables:     None
###
### Returns:       0 if error in field information (field not filled in or
###                  field data invalid (see globals below)
###                1 if successful (all fields filled in and valid)
###
### Uses Globals:  %VAR --  Miscellaneous variables associative array
###                         Sets $VAR{'TITLE'} with the title of an error message
###                         Sets $VAR{'MESSAGE'} with a message describing error
###                %FORM -- Form variables associative array
###
### Files Created: None

sub register { 
    local(%REQUIRED) = ('NAME', 'Name', 'EMAIL', 'E-Mail', 'EMAIL2', 'E-Mail Verification', 
                        'USER_ID', 'User ID', 'PASSWORD', 'Password', 
                        'PASSWORD2', 'Password Verification', 'VERIFY', 'Verify');
    local($missing, $len) = '';
    $VAR{'TITLE'} = '';
    $VAR{'MESSAGE'} = '';
    
    # Check to see if all the fields were filled in
    
    foreach $field (keys %REQUIRED) {
        if (!$FORM{$field}) { $missing .= "$REQUIRED{$field}, " }
    }
    if ($missing) {
        $VAR{'TITLE'} = 'MISSING INFORMATION';
        chop($missing); chop($missing);
        $VAR{'MESSAGE'} =  "You did not fill in the following information:<P>$missing";
    }                          
    
    # Check to see if the email addresses match
    
    elsif ($FORM{'EMAIL'} ne $FORM{'EMAIL2'}) {
        $VAR{'TITLE'} = 'EMAIL ADDRESS ERROR';
        $VAR{'MESSAGE'} =  'The two email addresses you entered do not match. ';
    }
        
    # Check to see if the email addresses are the correct format
    
    elsif (!&email_check($FORM{'EMAIL'})) {
        $VAR{'TITLE'} = 'EMAIL ADDRESS ERROR';
        $VAR{'MESSAGE'} =  'The email address you entered is not a valid address.';
    }
        
    # Check to see if the User ID is in the correct format
    
    elsif ((($len = length($FORM{'USER_ID'})) < 4) || ($len > 10)) {
        $VAR{'TITLE'} = 'USER ID ERROR';
        $VAR{'MESSAGE'} = 'Your User ID must be between 4 and 10 characters long.';
    }
    elsif ($FORM{'USER_ID'} =~ /[^\041-\176]/) {
        $VAR{'TITLE'} = 'USER ID ERROR';
        $VAR{'MESSAGE'} = 'The User ID you entered contains an invalid character.';
    }       
    
    # Check to see if the password fields match
    
    elsif ($FORM{'PASSWORD'} ne $FORM{'PASSWORD2'}) {
        $VAR{'TITLE'} = 'PASSWORD ERROR';
        $VAR{'MESSAGE'} = 'The two passwords you entered do not match. ';
    }

    # Check to see if the password is in the correct format
    
    elsif ((($len = length($FORM{'PASSWORD'})) < 6) || ($len > 10)) {
        $VAR{'TITLE'} = 'PASSWORD ERROR';
        $VAR{'MESSAGE'} = 'Your password must be between 6 and 10 characters long.';
    }
    elsif ($FORM{'PASSWORD'} =~ /[^\041-\176]/) {
        $VAR{'TITLE'} = 'PASSWORD ERROR';
        $VAR{'MESSAGE'} = 'The password you entered contains an invalid character.';
    } 
        
    # Return error if any problems
    
    if ($VAR{'TITLE'}) {
        $VAR{'MESSAGE'} .= '<P>Please press the back button on your browser to return ';
        $VAR{'MESSAGE'} .= 'to the registration page and make the necessary corrections.';
        return(0);
    }
    else { return(1); }
}

1;
